import json
import os

import requests

from .llm_provider import LLMProvider


class OllamaClient(LLMProvider):
    def __init__(self, primer=None):
        self.provider = "ollama"
        self.base_url = os.environ.get("OLLAMA_URL")
        self.models = self.list_models()  # List available models on initialization
        self.model = (
            self.models[0] if self.models else None
        )  # Set the first model as default
        self.messages = []
        self.primer = primer
        if self.primer:
            self.messages.append({"role": "system", "content": self.primer})

    def list_models(self):
        """List all models available from Ollama's /api/tags endpoint"""
        url = f"{self.base_url}/api/tags"
        response = requests.get(url)

        if response.status_code == 200:
            models_data = response.json()["models"]  # Extract the list of models
            return [
                model["name"] for model in models_data
            ]  # Return the 'name' of each model
        else:
            raise Exception(
                f"Failed to list models: {response.status_code} {response.text}"
            )

    def chat(self, message):
        """Chat with the model using Ollama's API"""
        self.messages.append({"role": "user", "content": message})
        current_message = ""

        # Update the URL to the correct chat endpoint
        url = f"{self.base_url}/api/chat"
        payload = {
            "model": self.model,  # The model to use for chat
            "messages": self.messages,  # The chat history including the primer
        }

        # Sending the chat request and streaming the response
        response_stream = requests.post(url, json=payload, stream=True)

        # Loop through the response lines (streamed chunks)
        for chunk in response_stream.iter_lines():
            if chunk:
                # Each chunk is a valid JSON object
                response_json = json.loads(chunk.decode("utf-8"))

                # Check if it's the final chunk with 'done' flag
                if response_json.get("done", False):
                    print("Final response:", response_json)  # For debugging purposes
                    break

                # Get the content of the message if available
                content = response_json.get("message", {}).get("content", "")
                current_message += content  # Append the content to the full message

                yield content  # Yield the content incrementally if needed

        # Add the final assistant message to the conversation history
        self.messages.append({"role": "assistant", "content": current_message})

        return current_message  # Return the final assembled message


if __name__ == "__main__": # pragma: no cover
    client = OllamaClient(primer="Limit your response to 300 characters or less")
    for text in client.chat(
        "I'm traveling to Madrid soon (mid-October) with my wife. We love food, history, and shopping. We've been there before. Can you recommend a few destinations/activities off the beaten path? We're staying in the city center and will be there for 3 days. We're looking for authentic experiences, not tourist traps."
    ):
        print(text, end="", flush=True)
    print()



""" Base class for LLM providers """
import os
from dotenv import load_dotenv
load_dotenv(os.path.expanduser("~/.config/.env" if os .path.exists(os.path.expanduser("~/.config")) else "~/.env"))


class LLMProvider:

    def __init__(self, primer=None, model=None):
        self.name = None
        self.provider = None

        self.messages = []
        self.primer = None

    def list_models(self):
        """ List all models available for this provider """
        return []
    
    def refresh_models(self):
        """ Refresh the list of models available for this provider """
        print("Not implemented for this provider")

    def chat(self, message):
        """ Chat with the LLM provider 
        return a generator that yields the response text 
        """
        pass
    
    @property
    def model(self):
        return self._model

    @model.setter
    def model(self, value):
        for m in self.models:
            if value in m:
                self._model = m
                return
        raise ValueError(f"Model {value} not found for provider {self.provider}")
    
    def new_chat(self, primer=None):
        """ Create a new chat (erase messages history)
        Note that the semantics of the primer is that we reuse the original primer
        when not specified here.
        """
        self.messages = []
        if primer:
            self.primer = primer
        if self.primer:
            # Default is to use OpenAI system prompt
            self.messages.append({"role": "system", "content": self.primer})

    def n_user_messages(self):
        """ Number of user messages in the history """
        return len([m for m in self.messages if m['role'] == 'user'])

    def __str__(self):
        return f"{self.provider}:{self.model}"
    
    def __repr__(self):
        return f"{self.provider}:{self.model}"


import os
import json


def load_config(filename="config.json"):
    if not os.path.exists(filename):
        return {}
    with open(filename, "r") as f:
        return json.load(f)
    

def default_provider(filename="config.json"):
    config = load_config(filename)
    return config.get("default-provider", "openai")


def default_model(filename="config.json", provider=None):
    config = load_config(filename)
    if provider is None:
        provider = default_provider(filename)
    return config.get('default-models', {}).get(provider, {}).get("model")
    

if __name__ == "__main__":
    config = load_config()
    print(json.dumps(config, indent=4))

    print("default provider:", default_provider())
    print("default model:", default_model())

    config_file = "does-not-exist.json"
    config = load_config(config_file)
    print(json.dumps(config, indent=4))

    print("default provider:", default_provider(config_file))
    print("default model:", default_model(config_file))
    



import io
import re
from unittest.mock import patch, call
from io import StringIO

from prompt_toolkit.input import create_pipe_input
import pytest
from prompt_toolkit.application import create_app_session
from prompt_toolkit.output import DummyOutput

from az.az import main

"""
I am using multiple methods to test the CLI. Just my knowledge of how to test prompt_toolkit apps 
which evolved over time.

TODO:
- mock configured providers (the reading of env vars should be done in main(), otherwise it's not mocked)
"""


class MockLLM:
    def __init__(self, response):
        self.response = response
        self._n_user_messages = 0

    def chat(self, prompt):
        # can simulate a list of responses as well as a single response
        if isinstance(self.response, list):
            for r in self.response: 
                yield r
        else:
            yield self.response
        self._n_user_messages += 1

    def n_user_messages(self):
        return self._n_user_messages
    
    def __str__(self):
        return "MockLLM"
    

@pytest.fixture(autouse=True, scope="function")
def modify_env_variables(monkeypatch):
    for var in ['OPENAI_API_KEY', 'OLLAMA_URL', 'ANTHROPIC_API_KEY', 'GEMINI_API_KEY']:
        monkeypatch.delenv(var, raising=False)
    yield


def test_cli_app():
    output_stream = io.StringIO()
    response = "This is a fixed mock response."
    mock_llm = MockLLM(response=response)
    
    with patch('sys.stdout', output_stream), \
         patch('az.az.provider_factory', return_value=mock_llm), \
         patch('prompt_toolkit.PromptSession.prompt', side_effect=["Test prompt", EOFError]), \
         patch('sys.argv', ['program_name']):
        
        main() 
        
        output = output_stream.getvalue()
        assert response in output
        assert "Bye" in output


def test_bullets_rendering():
    # Test that markdown rendering happens
    output_stream = io.StringIO()
    responses = [
        "This is the first mock response.",
        "This is the second mock response.",
        "This is the third mock response."
    ]

    mock_llm = MockLLM(response=[ f"- {response}\n\n" for response in responses]) 
    
    with patch('sys.stdout', output_stream), \
         patch('az.az.provider_factory', return_value=mock_llm), \
         patch('prompt_toolkit.PromptSession.prompt', side_effect=["Test prompt", EOFError]), \
         patch('sys.argv', ['program_name']):
        
        main() 
        
    output = output_stream.getvalue()
    print(output)
        
    for response in responses:
        assert response.rstrip() in output
    assert '•' in output


def test_help_message():
    output_stream = io.StringIO()
    mock_llm = MockLLM(response="") 
    
    with patch('sys.stdout', new=output_stream, create=True), \
         patch('az.az.provider_factory', return_value=mock_llm), \
         patch('prompt_toolkit.PromptSession.prompt', side_effect=["?\n", EOFError]), \
         patch('sys.argv', ['program_name']):
        
        main() 
        
    output = output_stream.getvalue()
    print(output)
        
    assert re.search(r'^.*Command.*Description.*$', output, re.MULTILINE)
    


def test_double_enter_does_not_call_chat_when_single_enter():
    output_stream = io.StringIO()

    input_stream = io.StringIO("twice\n\n\n\n")
    mock_llm = MockLLM(response="")
    
    with patch('sys.stdout', output_stream), \
         patch('az.az.provider_factory', return_value=mock_llm), \
         patch('sys.argv', ['program_name', '-d']), \
         create_pipe_input() as inp,  \
         patch.object(MockLLM, 'chat', wraps=mock_llm.chat) as mock_chat, \
         patch('az.az.get_input', return_value=inp):
        
        inp.send_text("test\n\nq\n")
        main() 

    output = output_stream.getvalue()
    print(output)

    # Assert that chat() is not called after single enter
    assert mock_chat.call_count == 1
        

@pytest.fixture(autouse=True, scope="function")
def mock_input():
    with create_pipe_input() as pipe_input:
        with create_app_session(input=pipe_input, output=DummyOutput()):
            yield pipe_input


def test_sample_app(mock_input):
    # TODO: not sure why I need two \n\n to get the prompt to show up
    mock_input.send_text("4\n\nq\n\n")
    response = "the response"
    mock_llm = MockLLM(response=response)
    out = StringIO()
    with patch('sys.stdout', new=out), \
        patch('sys.argv', ['program_name']), \
         patch('az.az.provider_factory', return_value=mock_llm):
        main()
    output = out.getvalue()
    print(f"out:\n\n{output}\n\n-----")
    assert response in output


@pytest.mark.skip(reason="real llm is slow. use this only as an exception, also it requires env vars")
def test_sample_app_real_llm(mock_input):
    mock_input.send_text("how much are 2 * 8?\n\nq\n\n")
    out = StringIO()
    with patch('sys.stdout', new=out), \
         patch('sys.argv', ['program_name']):
        main()
    output = out.getvalue()
    print(f"out:\n\n{output}\n\n-----")
    assert "16" in output






import json
import os
from typing import Dict, List, Set

class FileCache:
    def __init__(self, cache_file: str):
        self.cache_file = cache_file
        self.cache: Dict[str, Set[str]] = self._load_cache()

    def _load_cache(self) -> Dict[str, Set[str]]:
        if os.path.exists(self.cache_file):
            with open(self.cache_file, 'r') as f:
                data = json.load(f)
                return {k: set(v) for k, v in data.items()}
        return {}

    def _save_cache(self):
        with open(self.cache_file, 'w') as f:
            json.dump({k: list(v) for k, v in self.cache.items()}, f)

    def get(self, key: str) -> Set[str]:
        return self.cache.get(key, set())

    def set(self, key: str, value: List[str]):
        self.cache[key] = set(value)
        self._save_cache()

    def update(self, key: str, value: List[str]):
        self.cache.setdefault(key, set()).update(value)
        self._save_cache()

    def clear(self):
        self.cache.clear()
        self._save_cache()


if __name__ == "__main__": # pragma: no cover
    import random
    if os.path.exists("cache.json") and random.random() < 0.5:
        print("removing cache.json")
        os.remove("cache.json")
    cache = FileCache("cache.json")
    if len(cache.get("openai")) == 0:
        print("setting cache")
        cache.set("openai", ["gpt-4", "gpt-4o-mini"])
    else:
        print("cache hit")
        models = cache.get("openai")
        print("models:", models)
        
    # print(cache.get("openai"))








from setuptools import setup, find_packages

from az.utils import remove_images_from_markdown


# Read the content of README.md
with open('README.md', 'r') as f:
    readme_content = f.read()

# Remove images from the readme content
readme_content_without_images = remove_images_from_markdown(readme_content)

setup(
    name='azc',
    version='1.0.9',
    packages=find_packages(),
    install_requires=open('requirements.txt').read().splitlines(),
    author='Sagi Smolarski',
    author_email='sagism@gmail.com',
    description='Command-line tool for interacting with LLMs',
    long_description=readme_content_without_images,
    long_description_content_type='text/markdown',
    url='https://github.com/sagism/azc',
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6',
    py_modules=['az'],
    entry_points={
        'console_scripts': [
            'azc=az.az:main',
        ],
    }
)



from .llm_provider import LLMProvider
import anthropic



class AnthropicClient(LLMProvider):
    def __init__(self, config={}, primer=None):
        self.provider = 'anthropic'
        self.client = anthropic.Anthropic()
        self.models = self.list_models()
        self.config = config
        self.model = self.config.get("anthropic", {}).get("model", "claude-3-5-sonnet")
        self.messages = []
        self.primer = primer
          

    def list_models(self):
        """ No way to list models in anthropic? """
        return ["claude-3-5-sonnet-20240620"]
    

    def chat(self, message):
        if self.primer and len(self.messages) == 0:
            self.messages.append({"role": "user", "content": self.primer + "\n\n" + message})
        else:
            self.messages.append({"role": "user", "content": message})
        with self.client.messages.stream(
            max_tokens=1024,
            messages=self.messages,
            model=self.model,
                ) as stream:
            for text in stream.text_stream:
                yield text
        self.messages.append({"role": "assistant", "content": text})


    def new_chat(self, primer=None):
        self.messages = []
        if primer:
            self.primer = primer


if __name__ == "__main__": # pragma: no cover
    client = AnthropicClient(primer="Limit your response to 300 characters or less")
    for text in client.chat("Provide a list of 7 things I could do when I have a spare hour at home, which won't waste my time?"):
        print(text, end="", flush=True)
    print()
    print("---")
    
    for text in client.chat("out of those, which one would you recommend?"):
        print(text, end="", flush=True)
    print()


# chat_app.py

import sys
import os
import shutil
import argparse

# a mix of rich and prompt_toolkit seem to hit the sweet spot for terminal UI interactivity
from rich.console import Console
from rich.markdown import Markdown
from rich.panel import Panel
from rich.align import Align
from rich.live import Live
from rich.box import Box
from prompt_toolkit import PromptSession
from prompt_toolkit.patch_stdout import patch_stdout
from prompt_toolkit.formatted_text import HTML
from prompt_toolkit.history import FileHistory
from prompt_toolkit.completion import Completer, Completion
from prompt_toolkit.key_binding import KeyBindings


import readline # needed for prompt editing

from .utils import number_to_ordinal
from .config import load_config, default_model, default_provider

# Providers
from .ollama_provider import OllamaClient
from .openai_provider import OpenAIClient
from .anthropic_provider import AnthropicClient
from .gemini_provider import GeminiClient

HISTORY_FILE_NAME = os.path.expanduser("~/.config/.azc_history" if os.path.exists(os.path.expanduser("~/.config")) else "~/.azc_history")

config = load_config()

bindings = KeyBindings()

def is_command(string: str) -> bool:
    if string.strip().lower() in ('exit', 'quit', 'q', 'p', 'l', 'm', 'p', 'n', 'h', 'r', '?', ''): return True
    elif string.strip().startswith("p "): return True
    else: return False


def get_input():
    # added so we can override input for testing
    return None


# Bind 'Control+n' to insert a newline
@bindings.add('c-n')
def insert_newline(event):
    event.current_buffer.insert_text('\n')


# An empty box border makes it easier to copy and paste.
EMPTY: Box = Box(
    "    \n"
    "    \n"
    "    \n"
    "    \n"
    "    \n"
    "    \n"
    "    \n"
    "    \n",
    ascii=True,
)

# Initialize the console
console = Console()

# this is done by the providers so it's not needed here
# load_dotenv(os.path.expanduser("~/.config/.env" if os .path.exists(os.path.expanduser("~/.config")) else "~/.env"))


providers = []
if 'OPENAI_API_KEY' in os.environ:
    providers.append('openai')
if 'OLLAMA_URL' in os.environ:
    providers.append('ollama')
if 'ANTHROPIC_API_KEY' in os.environ:
    providers.append('anthropic')
if 'GEMINI_API_KEY' in os.environ:
    providers.append('gemini')

# provider_completer = WordCompleter([f"p {provider}" for provider in providers], ignore_case=True)

class CommandsCompleter(Completer):
    """
    This completer handles the 'p' command for changing provider
    """
    def get_completions(self, document, complete_event):
        text = document.current_line
        if text.startswith('p '):   
            for provider in providers:
                yield Completion(
                f'p {provider}', start_position=-1000,
                display=HTML(f'{provider}'),
                style='bg:ansiyellow')
        
completer = CommandsCompleter()


def primer():
    _, rows = shutil.get_terminal_size()
    # single screen is better...
    return f"please limit your response to {rows-4} lines at most)"


def provider_factory(provider_hint):
    provider_found = False

    for provider_name in providers:
        if provider_hint in provider_name:
            provider_full_name = provider_name
            provider_found = True
            break

    if not provider_found:
        raise ValueError(f"Cannot find provider with <{provider_hint}>")
    
    if provider_full_name == 'ollama':
        return OllamaClient(primer=primer())
    elif provider_full_name == 'openai':
        return OpenAIClient(config, primer=primer())
    elif provider_full_name == 'anthropic':
        return AnthropicClient(config, primer=primer())
    elif provider_full_name == 'gemini':
        return GeminiClient(config, primer=primer())



class FilteredHistory(FileHistory):
    """
    This class is a custom history class that filters out commands we don't want to save
    """
    def store_string(self, string: str) -> None:
        if not is_command(string):
                super().store_string(string)


def help():
    return """

Just type your message and press enter to start a chat.

| Command | Description |
|---------|-------------|
| l       | List models |
| r       | Refresh models |
| n       | New chat ( )   |
| ? or h  | Help (this screen) |
| m       | Change model |
| p provider_name | Change provider (p and space trigger autocomplete) |
| ctrl-n  | New line |
"""



def main(initial_prompt=None):
    if len(providers) == 0:
        console.print('no providers found, exiting, please set one of the following: OPENAI_API_KEY, OLLAMA_URL, ANTHROPIC_API_KEY, GEMINI_API_KEY in a .env file')
        return
    
    parser = argparse.ArgumentParser(description="Chat with an AI assistant")
    parser.add_argument("-p", "--provider", help="The provider to use, e.g. 'openai' or 'ollama'. Abbreviations allowed, like 'op' for 'openai'")
    parser.add_argument("-m", "--model", help="The model to use")
    parser.add_argument("-d", "--double-enter", action="store_true", help="Enable 'press enter twice to submit' mode")
    parser.add_argument("-b", "--batch", action="store_true", help="Not interactive, just do one chat and exit")
    parser.add_argument("-v", "--verbose", action="store_true", help="Verbose output")
    parser.add_argument("initial_prompt", nargs='?', help="Initial prompt (if provided without a flag)")
    args = parser.parse_args()
    initial_prompt = args.initial_prompt

    if args.verbose:
        console.print('providers configured: [yellow]' + ', '.join(providers) + '[/]')

    if args.double_enter:
        console.print("type <enter> twice to submit...")
        @bindings.add('enter')
        def _(event):
            buffer = event.current_buffer
            accepted = False
            if buffer.document.text.strip() == '':
                accepted = True
            elif is_command(buffer.document.text):
                accepted = True
            elif buffer.document.text.endswith('\n') and buffer.document.is_cursor_at_the_end:
                accepted = True
            else:
                accepted = False
            if accepted:
                buffer.validate_and_handle()
            else:
                buffer.insert_text('\n')



    provider_name = args.provider if args.provider else default_provider()
    provider_name = provider_name if provider_name else providers[0]
    
    client = provider_factory(provider_name)

    if args.model:
        client.model = args.model
    
    if args.verbose:
        console.print(f'using: [green]{client}[/]')
        console.print('[magenta]type ? or h for help[/]')

    def bottom_toolbar():
        return HTML(f' Using <b>{client}</b> ({number_to_ordinal(client.n_user_messages()+1)} message)     <ansicyan>enter ? or h for help</ansicyan> ')

    our_history = FilteredHistory(HISTORY_FILE_NAME)
    session = PromptSession(history=our_history, input=get_input())

    done=False
    
    try:
        while not done:
            # If there's an initial prompt, process it first
            if initial_prompt:
                user_input = initial_prompt
                initial_prompt = None
                if args.batch:
                    done=True
            else:
                # Get user input using prompt_toolkit
                with patch_stdout():
                    try:
                        user_input = session.prompt(
                            HTML(f'<ansicyan>azc></ansicyan> '),
                            completer=completer,
                            bottom_toolbar=bottom_toolbar,
                            key_bindings=bindings
                        )
                    except EOFError:
                        done = True
                        break


            if user_input.strip().lower() == '':
                # Some people like to press enter to get a new line
                continue

            if user_input.strip().lower() in ('exit', 'quit', 'q'):
                break

            if user_input.strip().lower() in ('l'):
                console.print(Markdown('  - ' + '\n  - '.join(client.list_models())))
                continue

            if user_input.strip().lower() in ('n'):
                client.new_chat()
                continue

            if user_input.strip().lower() in ('r'):
                client.refresh_models()
                continue

            if user_input.strip().lower() in ('h', '?'):
                console.print(Markdown(help()))
                continue

            if user_input.strip().lower() in ('m'):
                model = session.prompt(
                    HTML(f'<ansicyan>model (partial name okay): </ansicyan> '),
                    multiline=False,
                    is_password=False
                )
                client.model = model
                continue

            if user_input.startswith('p '):
                provider_name = user_input.split(' ')[1]
                try:
                    client = provider_factory(provider_name)
                    console.print(f'using: {client} (new chat)')
                except ValueError as e:
                    console.print(f'[red]error: {e}[/]')
                continue

            title = f"{client} ({number_to_ordinal(client.n_user_messages()+1)} message)" if not args.batch else None

            assistant_panel = Panel(
                Align.left(""),
                title=title,
                style="yellow",
                expand=True,
                box=EMPTY
            )

            current_message = ""

            # Note that vertical_overflow="visible" causes realtime updates of rendered markdown beyond the full window height,
            # but it leaves a trail of partially rendered markdown behind when new content is added, hence it is not used
            with Live(assistant_panel, console=console, refresh_per_second=2, vertical_overflow='ellipsis') as live:
                if args.double_enter:
                    console.print(f'...')
                for chunk in client.chat(user_input):
                    current_message += chunk
                    new_panel = Panel(
                        Align.left(Markdown(current_message)),
                        title=title,
                        style="yellow",
                        expand=True,
                        border_style="none",
                        box=EMPTY
                    )
                    live.update(new_panel, refresh=False)

    except KeyboardInterrupt:
        done = True
    finally:
        if not args.batch:
            console.print(":wave: [italic]Bye[/]")

if __name__ == "__main__": # pragma: no cover
    main()





import os

from .llm_provider import LLMProvider
import google.generativeai as genai


genai.configure(api_key=os.environ.get("GEMINI_API_KEY"))

class GeminiClient(LLMProvider):
    def __init__(self, config={}, primer=None):
        self.primer = primer
        self.provider = 'gemini'
        self.config = config

        self.models = self.list_models()
        self.model = self.config.get(self.provider, {}).get("model", "gemini-1.5-flash")
        
        self._n_user_messages = 0
        self.client = genai.GenerativeModel(
            model_name=self.model,
            system_instruction=self.primer
        )        
        self._chat = self.client.start_chat()
          

    def list_models(self):
        # I don't see a way to list models in the API
        return ['gemini-1.5-flash', 'gemini-1.5-pro']

    def n_user_messages(self):
        return self._n_user_messages

    def chat(self, message):
        current_message = ""
        response_stream = self._chat.send_message(message, stream=True)
        for chunk in response_stream:
            delta = chunk.text
            current_message += delta
            yield delta

        self._n_user_messages += 1
        return current_message

    def new_chat(self, primer=None):
        if primer:
            self.primer = primer
        self._chat = self.client.start_chat()
        self._n_user_messages = 0


if __name__ == "__main__": # pragma: no cover
    client = GeminiClient(primer="Limit your response to 300 characters or less")
    for text in client.chat("I'm traveling to Madrid soon (mid-October) with my wife. We love food, history and shopping. We've been there before. Can you recommend a few destinations/activities off the beaten path? We're staying in the city center and will be there for 3 days. We're looking for authentic experiences, not tourist traps."):
        print(text, end="", flush=True)
    print()

    print(client.list_models())
